//
//  FMFMapView.m
//  map4d_map
//
//  Created by Sua Le on 05/05/2021.
//


#import "FMFMapView.h"
#import "Map4dFLTConvert.h"
#import "Map4dFLTMethod.h"

#define kMFMinZoomLevel 2
#define kMFMaxZoomLevel 22

#pragma mark - FMFMapViewFactory

@implementation FMFMapViewFactory {
  NSObject<FlutterPluginRegistrar>* _registrar;
}

- (instancetype)initWithRegistrar:(NSObject<FlutterPluginRegistrar>*)registrar {
  self = [super init];
  if (self) {
    _registrar = registrar;
  }
  return self;
}

- (NSObject<FlutterMessageCodec>*)createArgsCodec {
  return [FlutterStandardMessageCodec sharedInstance];
}

- (NSObject<FlutterPlatformView>*)createWithFrame:(CGRect)frame
                                   viewIdentifier:(int64_t)viewId
                                        arguments:(id _Nullable)args {
  return [[FMFMapView alloc] initWithFrame:frame
                            viewIdentifier:viewId
                                 arguments:args
                                 registrar:_registrar];
}
@end

#pragma mark - Event tracking configurations

@interface FMFEventTracking : NSObject
@property (nonatomic) BOOL cameraPosition;
@end
@implementation FMFEventTracking
- (instancetype)init {
  self = [super init];
  if (self) {
    _cameraPosition = NO;
  }
  return self;
}
@end


#pragma mark - FMFMapView

@implementation FMFMapView {
  MFMapView* _mapView;
  int64_t _viewId;
  FlutterMethodChannel* _channel;
  NSObject<FlutterPluginRegistrar>* _registrar;
  FMFEventTracking* _track;
}

- (instancetype)initWithFrame:(CGRect)frame
               viewIdentifier:(int64_t)viewId
                    arguments:(id _Nullable)args
                    registrar:(NSObject<FlutterPluginRegistrar>*)registrar {
  if (self = [super init]) {
    _viewId = viewId;
    _registrar = registrar;
    _track = [[FMFEventTracking alloc] init];
    
    // setup MFMapView
    _mapView = [[MFMapView alloc] initWithFrame:frame];
    _mapView.delegate = self;
    [self setupMapViewWithArguments:args];
    
    // flutter channel
    NSString* channelName = [NSString stringWithFormat:@"plugin:map4d-map-view-type_%lld", viewId];
    _channel = [FlutterMethodChannel methodChannelWithName:channelName
                                           binaryMessenger:registrar.messenger];
    __weak __typeof__(self) weakSelf = self;
    [_channel setMethodCallHandler:^(FlutterMethodCall * call, FlutterResult result) {
      if (weakSelf) {
        [weakSelf onMethodCall:call result:result];
      }
    }];
  }
  return self;
}

- (void)setupMapViewWithArguments:(id _Nullable)args {
  // initial map options
  [self interpretMapOptions:args[@"options"]];
  
  // initial camera position
  MFCameraPosition* camera = [Map4dFLTConvert toCameraPosition:args[@"initialCameraPosition"]];
  if (camera != nil) {
    _mapView.camera = camera;
  }
}

- (UIView*)view {
  return _mapView;
}

// Method call handler
- (void)onMethodCall:(FlutterMethodCall*)call result:(FlutterResult)result {
  FMFMethodID methodID = [[Map4dFLTMethod shared] getIdByName:call.method];
  switch (methodID) {
    /* map # update **/
    case FMFMethodMapUpdate:
      [self interpretMapOptions:call.arguments[@"options"]];
      result(nil);
      break;

    /* map # get zoom level **/
    case FMFMethodGetZoomLevel:
      result(@(_mapView.camera.zoom));
      break;

    /* map # get camera for bounds **/
    case FMFMethodCameraForBounds: {
      MFCoordinateBounds* bounds = [Map4dFLTConvert toCoordinateBounds:call.arguments[@"bounds"]];
      double padding = [Map4dFLTConvert toDouble:call.arguments[@"padding"]];
      UIEdgeInsets insets = UIEdgeInsetsMake(padding, padding, padding, padding);
      MFCameraPosition* camera = [_mapView cameraForBounds:bounds insets:insets];
      result([Map4dFLTConvert positionToJson:camera]);
      break;
    }

    /* map # fit bounds **/
    case FMFMethodFitBounds: {
      MFCoordinateBounds* bounds = [Map4dFLTConvert toCoordinateBounds:call.arguments[@"bounds"]];
      double padding = [Map4dFLTConvert toDouble:call.arguments[@"padding"]];
      UIEdgeInsets insets = UIEdgeInsetsMake(padding, padding, padding, padding);
      [_mapView moveCamera:[MFCameraUpdate fitBounds:bounds withEdgeInsets:insets]];
      result(nil);
      break;
    }
      
    /* map # get camera position **/
    case FMFMethodGetCameraPosition: {
      if (_mapView != nil) {
        result([Map4dFLTConvert positionToJson:_mapView.camera]);
      } else {
        result([FlutterError errorWithCode:@"MFMapView uninitialized"
                                   message:@"getCameraPosition called prior to map initialization"
                                   details:nil]);
      }
      break;
    }

    /* map # set 3D mode **/
    case FMFMethodEnable3DMode: {
      BOOL isEnable = [Map4dFLTConvert toBool:call.arguments[@"enable3DMode"]];
      [_mapView enable3DMode: isEnable];
      result(nil);
      break;
    }

    /* camera # move **/
    case FMFMethodMoveCamera: {
      MFCameraUpdate* cameraUpdate = [Map4dFLTConvert toCameraUpdate:call.arguments[@"cameraUpdate"]];
      [_mapView moveCamera:cameraUpdate];
      result(nil);
      break;
    }

    /* camera # animate **/
    case FMFMethodAnimateCamera: {
      MFCameraUpdate* cameraUpdate = [Map4dFLTConvert toCameraUpdate:call.arguments[@"cameraUpdate"]];
      [_mapView animateCamera:cameraUpdate];
      result(nil);
      break;
    }

    /* default **/
    default:
      NSLog(@"Unknow call method: %@", call.method);
      result(nil);
      break;
  }
}

#pragma mark - Interpret Map options
- (void)interpretMapOptions:(NSDictionary*)data {
  //  NSArray* cameraTargetBounds = data[@"cameraTargetBounds"];
  //  if (cameraTargetBounds) {
  //    [sink setCameraTargetBounds:ToOptionalBounds(cameraTargetBounds)];
  //  }
  
  id buildingsEnabled = data[@"buildingsEnabled"];
  if (buildingsEnabled) {
    [self setBuildingsEnabled:[Map4dFLTConvert toBool:buildingsEnabled]];
  }
  
  id poisEnabled = data[@"poisEnabled"];
  if (poisEnabled) {
    [self setPOIsEnabled:[Map4dFLTConvert toBool:poisEnabled]];
  }
  
  NSArray* zoomData = data[@"minMaxZoomPreference"];
  if (zoomData) {
    float minZoom = (zoomData[0] == [NSNull null]) ? kMFMinZoomLevel : [Map4dFLTConvert toFloat:zoomData[0]];
    float maxZoom = (zoomData[1] == [NSNull null]) ? kMFMaxZoomLevel : [Map4dFLTConvert toFloat:zoomData[1]];
    [self setMinZoom:minZoom maxZoom:maxZoom];
  }
  
  NSNumber* rotateGesturesEnabled = data[@"rotateGesturesEnabled"];
  if (rotateGesturesEnabled != nil) {
    [self setRotateGesturesEnabled:[Map4dFLTConvert toBool:rotateGesturesEnabled]];
  }
  NSNumber* scrollGesturesEnabled = data[@"scrollGesturesEnabled"];
  if (scrollGesturesEnabled != nil) {
    [self setScrollGesturesEnabled:[Map4dFLTConvert toBool:scrollGesturesEnabled]];
  }
  NSNumber* tiltGesturesEnabled = data[@"tiltGesturesEnabled"];
  if (tiltGesturesEnabled != nil) {
    [self setTiltGesturesEnabled:[Map4dFLTConvert toBool:tiltGesturesEnabled]];
  }
  NSNumber* trackCameraPosition = data[@"trackCameraPosition"];
  if (trackCameraPosition != nil) {
    [self setTrackCameraPosition:[Map4dFLTConvert toBool:trackCameraPosition]];
  }
  NSNumber* zoomGesturesEnabled = data[@"zoomGesturesEnabled"];
  if (zoomGesturesEnabled != nil) {
    [self setZoomGesturesEnabled:[Map4dFLTConvert toBool:zoomGesturesEnabled]];
  }
  NSNumber* myLocationEnabled = data[@"myLocationEnabled"];
  if (myLocationEnabled != nil) {
    [self setMyLocationEnabled:[Map4dFLTConvert toBool:myLocationEnabled]];
  }
  NSNumber* myLocationButtonEnabled = data[@"myLocationButtonEnabled"];
  if (myLocationButtonEnabled != nil) {
    [self setMyLocationButtonEnabled:[Map4dFLTConvert toBool:myLocationButtonEnabled]];
  }
  NSNumber* mode3dEnabled = data[@"mode3dEnabled"];
  if (mode3dEnabled != nil) {
    [self set3DModeEnabled:[Map4dFLTConvert toBool:mode3dEnabled]];
  }
  NSNumber* waterEffectEnabled = data[@"waterEffectEnabled"];
  if (waterEffectEnabled != nil) {
    [self setWaterEffectEnabled:[Map4dFLTConvert toBool:waterEffectEnabled]];
  }
}

#pragma mark - FMFMapViewOptionsSink methods

- (void)setCamera:(MFCameraPosition*)camera {
  _mapView.camera = camera;
}

- (void)setCameraTargetBounds:(MFCoordinateBounds*)bounds {
  //TODO
  //  _mapView.cameraTargetBounds = bounds;
}

- (void)setBuildingsEnabled:(BOOL)enabled {
  [_mapView setBuildingsEnabled:enabled];
}

- (void)setPOIsEnabled:(BOOL)enabled {
  [_mapView setPOIsEnabled:enabled];
}

- (void)setMinZoom:(float)minZoom maxZoom:(float)maxZoom {
  [_mapView setMinZoom:minZoom maxZoom:maxZoom];
}

- (void)setRotateGesturesEnabled:(BOOL)enabled {
  _mapView.settings.rotateGestures = enabled;
}

- (void)setScrollGesturesEnabled:(BOOL)enabled {
  _mapView.settings.scrollGestures = enabled;
}

- (void)setTiltGesturesEnabled:(BOOL)enabled {
  _mapView.settings.tiltGestures = enabled;
}

- (void)setZoomGesturesEnabled:(BOOL)enabled {
  _mapView.settings.zoomGestures = enabled;
}

- (void)setMyLocationEnabled:(BOOL)enabled {
  [_mapView setMyLocationEnabled:enabled];
}

- (void)setMyLocationButtonEnabled:(BOOL)enabled {
  _mapView.settings.myLocationButton = enabled;
}

- (void)set3DModeEnabled:(BOOL)enabled {
  [_mapView enable3DMode:enabled];
}

- (void)setWaterEffectEnabled:(BOOL)enabled {
  [_mapView enableWaterEffect:enabled];
}

- (void)setTrackCameraPosition:(BOOL)enabled {
  _track.cameraPosition = enabled;
}

#pragma mark - MFMapViewDelegate

- (void)mapView: (MFMapView*)  mapView willMove: (BOOL) gesture {
  [_channel invokeMethod:@"camera#onMoveStarted" arguments:@{@"isGesture" : @(gesture)}];
}

- (void)mapView: (MFMapView*)  mapView movingCameraPosition: (MFCameraPosition*) position {
  if (_track.cameraPosition) {
    NSDictionary* response = [Map4dFLTConvert positionToJson:position];
    [_channel invokeMethod:@"camera#onMove" arguments:@{@"position" : response}];
  }
}

//- (void)mapView: (MFMapView*)  mapView didChangeCameraPosition:(MFCameraPosition*) position {}

- (void)mapView: (MFMapView*)  mapView idleAtCameraPosition: (MFCameraPosition *) position {
  [_channel invokeMethod:@"camera#onIdle" arguments:@{}];
}

- (void)mapView: (MFMapView*) mapView didTapAtCoordinate: (CLLocationCoordinate2D) coordinate {
  NSArray* response = [Map4dFLTConvert locationToJson:coordinate];
  [_channel invokeMethod:@"map#didTapAtCoordinate" arguments:@{@"coordinate": response}];
}

- (void)mapView: (MFMapView*) mapView onModeChange: (bool) is3DMode {
  [_channel invokeMethod:@"map#onModeChange" arguments:@{@"is3DMode": @(is3DMode)}];
}

///* Called after a base map building has been tapped */
- (void)mapView:(MFMapView*)mapView didTapBuildingWithBuildingID:(NSString*)buildingID name:(NSString*)name location:(CLLocationCoordinate2D)location {
  NSDictionary* arguments = @{
    @"buildingId": buildingID,
    @"name": name,
    @"location": [Map4dFLTConvert locationToJson:location]
  };
  [_channel invokeMethod:@"map#onTapBuilding" arguments:arguments];
}

///* Called after a base map POI has been tapped */
- (void)mapView: (MFMapView*)  mapView didTapPOIWithPlaceID: (NSString*) placeID name: (NSString*) name location: (CLLocationCoordinate2D) location {
  NSDictionary* arguments = @{
    @"placeId": placeID,
    @"name": name,
    @"location": [Map4dFLTConvert locationToJson:location]
  };
  [_channel invokeMethod:@"map#onTapPOI" arguments:arguments];
}

//- (void)mapView: (MFMapView*)  mapView didTapMyLocation: (CLLocationCoordinate2D) location;
//
//- (BOOL)shouldChangeMapModeForMapView: (MFMapView*)  mapView;
//- (BOOL)didTapMyLocationButtonForMapView: (MFMapView*) mapView;
//- (UIView *) mapView: (MFMapView *) mapView markerInfoWindow: (MFMarker *) marker;

@end
